@@ -1,5 +1,7 @@
 /******************************************************************************
- * Copyright (c) 2005 Eric Petit
+ * $Id: tracker.c 400 2006-06-20 00:28:16Z titer $
+ *
+ * Copyright (c) 2005-2006 Transmission authors and contributors
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -41,10 +43,17 @@ struct tr_tracker_s
     uint64_t       dateOk;
 
 #define TC_STATUS_IDLE    1
-#define TC_STATUS_CONNECT 2
-#define TC_STATUS_RECV    4
+#define TC_STATUS_RESOLVE 2
+#define TC_STATUS_CONNECT 4
+#define TC_STATUS_RECV    8
     char           status;
 
+#define TC_ATTEMPT_NOREACH 1
+#define TC_ATTEMPT_ERROR   2
+#define TC_ATTEMPT_OK      4
+    char           lastAttempt;
+
+    tr_resolve_t * resolve;
     int            socket;
     uint8_t      * buf;
     int            size;
@@ -52,31 +61,39 @@ struct tr_tracker_s
 
     int            bindPort;
     int            newPort;
+
+    uint64_t       download;
+    uint64_t       upload;
 };
 
 static void sendQuery  ( tr_tracker_t * tc );
 static void recvAnswer ( tr_tracker_t * tc );
 
-tr_tracker_t * tr_trackerInit( tr_handle_t * h, tr_torrent_t * tor )
+tr_tracker_t * tr_trackerInit( tr_torrent_t * tor )
 {
     tr_tracker_t * tc;
 
     tc           = calloc( 1, sizeof( tr_tracker_t ) );
     tc->tor      = tor;
-    tc->id       = h->id;
+    tc->id       = tor->id;
 
     tc->started  = 1;
 
+    tc->interval = 300;
     tc->seeders  = -1;
     tc->leechers = -1;
 
     tc->status   = TC_STATUS_IDLE;
+    tc->lastAttempt = TC_ATTEMPT_NOREACH;
     tc->size     = 1024;
     tc->buf      = malloc( tc->size );
 
-    tc->bindPort = h->bindPort;
+    tc->bindPort = *(tor->bindPort);
     tc->newPort  = -1;
 
+    tc->download = tor->downloaded;
+    tc->upload   = tor->uploaded;
+
     return tc;
 }
 
@@ -84,8 +101,18 @@ static int shouldConnect( tr_tracker_t * tc )
 {
     uint64_t now = tr_date();
 
-    /* In any case, always wait 5 seconds between two requests */
-    if( now < tc->dateTry + 5000 )
+    /* Unreachable tracker, try 10 seconds before trying again */
+    if( tc->lastAttempt == TC_ATTEMPT_NOREACH &&
+        now < tc->dateTry + 10000 )
+    {
+        return 0;
+    }
+
+    /* The tracker rejected us (like 4XX code, unauthorized IP...),
+       don't hammer it - we'll probably get the same answer next time
+       anyway */
+    if( tc->lastAttempt == TC_ATTEMPT_ERROR &&
+        now < tc->dateTry + 1000 * tc->interval )
     {
         return 0;
     }
@@ -136,25 +163,7 @@ int tr_trackerPulse( tr_tracker_t * tc )
 
     if( ( tc->status & TC_STATUS_IDLE ) && shouldConnect( tc ) )
     {
-        struct in_addr addr;
-
-        if( tr_fdSocketWillCreate( tor->fdlimit, 1 ) )
-        {
-            return 0;
-        }
-
-        if( tr_netResolve( inf->trackerAddress, &addr ) )
-        {
-            tr_fdSocketClosed( tor->fdlimit, 1 );
-            return 0;
-        }
-
-        tc->socket = tr_netOpen( addr, htons( inf->trackerPort ) );
-        if( tc->socket < 0 )
-        {
-            tr_fdSocketClosed( tor->fdlimit, 1 );
-            return 0;
-        }
+        tc->resolve = tr_netResolveInit( inf->trackerAddress );
 
         tr_inf( "Tracker: connecting to %s:%d (%s)",
                 inf->trackerAddress, inf->trackerPort,
@@ -164,10 +173,48 @@ int tr_trackerPulse( tr_tracker_t * tc )
                     ( 0 < tc->newPort ? "sending 'stopped' to change port" :
                       "getting peers" ) ) ) );
 
-        tc->status  = TC_STATUS_CONNECT;
+        tc->status  = TC_STATUS_RESOLVE;
         tc->dateTry = tr_date();
     }
 
+    if( tc->status & TC_STATUS_RESOLVE )
+    {
+        int ret;
+        struct in_addr addr;
+
+        ret = tr_netResolvePulse( tc->resolve, &addr );
+        if( ret == TR_RESOLVE_WAIT )
+        {
+            return 0;
+        }
+        else
+        {
+            tr_netResolveClose( tc->resolve );
+        }
+        
+        if( ret == TR_RESOLVE_ERROR )
+        {
+            tc->status = TC_STATUS_IDLE;
+            return 0;
+        }
+
+        if( tr_fdSocketWillCreate( tor->fdlimit, 1 ) )
+        {
+            tc->status = TC_STATUS_IDLE;
+            return 0;
+        }
+
+        tc->socket = tr_netOpen( addr, htons( inf->trackerPort ) );
+        if( tc->socket < 0 )
+        {
+            tr_fdSocketClosed( tor->fdlimit, 1 );
+            tc->status = TC_STATUS_IDLE;
+            return 0;
+        }
+
+        tc->status = TC_STATUS_CONNECT;
+    }
+
     if( tc->status & TC_STATUS_CONNECT )
     {
         /* We are connecting to the tracker. Try to send the query */
@@ -230,7 +277,11 @@ void tr_trackerClose( tr_tracker_t * tc )
 {
     tr_torrent_t * tor = tc->tor;
 
-    if( tc->status > TC_STATUS_IDLE )
+    if( tc->status == TC_STATUS_RESOLVE )
+    {
+        tr_netResolveClose( tc->resolve );
+    }
+    else if( tc->status > TC_STATUS_RESOLVE )
     {
         tr_netClose( tc->socket );
         tr_fdSocketClosed( tor->fdlimit, 1 );
@@ -247,21 +298,34 @@ static void sendQuery( tr_tracker_t * tc )
     char     * event;
     uint64_t   left;
     int        ret;
+    uint64_t   down;
+    uint64_t   up;
 
-    if( tc->started && 0 < tc->newPort )
-    {
-        tc->bindPort = tc->newPort;
-        tc->newPort = -1;
-    }
-
+    down = tor->downloaded - tc->download;
+    up = tor->uploaded - tc->upload;
     if( tc->started )
+    {
         event = "&event=started";
+        down = up = 0;
+        
+        if( 0 < tc->newPort )
+        {
+            tc->bindPort = tc->newPort;
+            tc->newPort = -1;
+        }
+    }
     else if( tc->completed )
+    {
         event = "&event=completed";
+    }
     else if( tc->stopped || 0 < tc->newPort )
+    {
         event = "&event=stopped";
+    }
     else
+    {
         event = "";
+    }
 
     left = tr_cpLeftBytes( tor->completion );
 
@@ -270,9 +334,9 @@ static void sendQuery( tr_tracker_t * tc )
             "info_hash=%s&"
             "peer_id=%s&"
             "port=%d&"
-            "uploaded=%lld&"
-            "downloaded=%lld&"
-            "left=%lld&"
+            "uploaded=%"PRIu64"&"
+            "downloaded=%"PRIu64"&"
+            "left=%"PRIu64"&"
             "compact=1&"
             "numwant=50&"
             "key=%s"
@@ -282,7 +346,7 @@ static void sendQuery( tr_tracker_t * tc )
             "User-Agent: Transmission/%d.%d\r\n"
             "Connection: close\r\n\r\n",
             inf->trackerAnnounce, tor->hashString, tc->id,
-            tc->bindPort, tor->uploaded, tor->downloaded,
+            tc->bindPort, up, down,
             left, tor->key, event, inf->trackerAddress,
             VERSION_MAJOR, VERSION_MINOR );
 
@@ -310,6 +374,8 @@ static void recvAnswer( tr_tracker_t * tc )
     int i;
     benc_val_t   beAll;
     benc_val_t * bePeers, * beFoo;
+    uint8_t * body;
+    int bodylen;
 
     if( tc->pos == tc->size )
     {
@@ -338,34 +404,62 @@ static void recvAnswer( tr_tracker_t * tc )
     tc->status  = TC_STATUS_IDLE;
     tc->dateTry = tr_date();
 
-    if( tc->pos < 1 )
+    if( tc->pos < 12 || ( 0 != memcmp( tc->buf, "HTTP/1.0 ", 9 ) &&
+                          0 != memcmp( tc->buf, "HTTP/1.1 ", 9 ) ) )
     {
-        /* We got nothing */
+        /* We don't have a complete HTTP status line */
+        tr_inf( "Tracker: incomplete HTTP status line" );
+        tc->lastAttempt = TC_ATTEMPT_NOREACH;
         return;
     }
 
-    /* Find the beginning of the dictionary */
-    for( i = 0; i < tc->pos - 18; i++ )
+    if( '2' != tc->buf[9] )
     {
-        /* Hem */
-        if( !memcmp( &tc->buf[i], "d8:interval", 11 ) ||
-            !memcmp( &tc->buf[i], "d8:complete", 11 ) ||
-            !memcmp( &tc->buf[i], "d14:failure reason", 18 ) )
+        /* we didn't get a 2xx status code */
+        tr_err( "Tracker: invalid HTTP status code: %c%c%c",
+                tc->buf[9], tc->buf[10], tc->buf[11] );
+        tc->lastAttempt = TC_ATTEMPT_ERROR;
+        return;
+    }
+
+    /* find the end of the http headers */
+    body = tr_memmem( tc->buf, tc->pos, "\015\012\015\012", 4 );
+    if( NULL != body )
+    {
+        body += 4;
+    }
+    /* hooray for trackers that violate the HTTP spec */
+    else if( NULL != ( body = tr_memmem( tc->buf, tc->pos, "\015\015", 2 ) ) ||
+             NULL != ( body = tr_memmem( tc->buf, tc->pos, "\012\012", 2 ) ) )
+    {
+        body += 2;
+    }
+    else
+    {
+        tr_err( "Tracker: could not find end of HTTP headers" );
+        tc->lastAttempt = TC_ATTEMPT_NOREACH;
+        return;
+    }
+    bodylen = tc->pos - (body - tc->buf);
+
+    /* Find and load the dictionary */
+    for( i = 0; i < bodylen; i++ )
+    {
+        if( !tr_bencLoad( &body[i], bodylen - i, &beAll, NULL ) )
         {
             break;
         }
     }
 
-    if( i >= tc->pos - 18 )
+    if( i >= bodylen )
     {
-        tr_err( "Tracker: no dictionary in answer" );
-        // printf( "%s\n", tc->buf );
-        return;
-    }
-
-    if( tr_bencLoad( &tc->buf[i], &beAll, NULL ) )
-    {
-        tr_err( "Tracker: error parsing bencoded data" );
+        if( tc->stopped || 0 < tc->newPort )
+        {
+            tc->lastAttempt = TC_ATTEMPT_OK;
+            goto nodict;
+        }
+        tr_err( "Tracker: no valid dictionary found in answer" );
+        tc->lastAttempt = TC_ATTEMPT_ERROR;
         return;
     }
 
@@ -374,13 +468,15 @@ static void recvAnswer( tr_tracker_t * tc )
     if( ( bePeers = tr_bencDictFind( &beAll, "failure reason" ) ) )
     {
         tr_err( "Tracker: %s", bePeers->val.s.s );
-        tor->status |= TR_TRACKER_ERROR;
-        snprintf( tor->error, sizeof( tor->error ),
+        tor->error |= TR_ETRACKER;
+        snprintf( tor->trackerError, sizeof( tor->trackerError ),
                   "%s", bePeers->val.s.s );
+        tc->lastAttempt = TC_ATTEMPT_ERROR;
         goto cleanup;
     }
 
-    tor->status &= ~TR_TRACKER_ERROR;
+    tor->error &= ~TR_ETRACKER;
+    tc->lastAttempt = TC_ATTEMPT_OK;
 
     if( !tc->interval )
     {
@@ -417,6 +513,10 @@ static void recvAnswer( tr_tracker_t * tc )
 
     if( !( bePeers = tr_bencDictFind( &beAll, "peers" ) ) )
     {
+        if( tc->stopped || 0 < tc->newPort )
+        {
+            goto nodict;
+        }
         tr_err( "Tracker: no \"peers\" field" );
         goto cleanup;
     }
@@ -477,6 +577,7 @@ static void recvAnswer( tr_tracker_t * tc )
         }
     }
 
+nodict:
     /* Success */
     tc->started   = 0;
     tc->completed = 0;
@@ -489,7 +590,9 @@ static void recvAnswer( tr_tracker_t * tc )
     }
     else if( 0 < tc->newPort )
     {
-        tc->started = 1;
+        tc->started  = 1;
+        tc->download = tor->downloaded;
+        tc->upload   = tor->uploaded;
     }
 
 cleanup:
@@ -506,6 +609,7 @@ int tr_trackerScrape( tr_torrent_t * tor, int * seeders, int * leechers )
     struct in_addr addr;
     uint64_t date;
     int pos, len;
+    tr_resolve_t * resolve;
 
     if( !tor->scrape[0] )
     {
@@ -513,10 +617,25 @@ int tr_trackerScrape( tr_torrent_t * tor, int * seeders, int * leechers )
         return 1;
     }
 
-    if( tr_netResolve( inf->trackerAddress, &addr ) )
+    resolve = tr_netResolveInit( inf->trackerAddress );
+    for( date = tr_date();; )
     {
-        return 0;
+        ret = tr_netResolvePulse( resolve, &addr );
+        if( ret == TR_RESOLVE_OK )
+        {
+            tr_netResolveClose( resolve );
+            break;
+        }
+        if( ret == TR_RESOLVE_ERROR ||
+            ( ret == TR_RESOLVE_WAIT && tr_date() > date + 10000 ) )
+        {
+            fprintf( stderr, "Could not resolve %s\n", inf->trackerAddress );
+            tr_netResolveClose( resolve );
+            return 1;
+        }
+        tr_wait( 10 );
     }
+
     s = tr_netOpen( addr, htons( inf->trackerPort ) );
     if( s < 0 )
     {
@@ -597,7 +716,7 @@ int tr_trackerScrape( tr_torrent_t * tor, int * seeders, int * leechers )
     {
         return 1;
     }
-    if( tr_bencLoad( &buf[i], &scrape, NULL ) )
+    if( tr_bencLoad( &buf[i], pos - i, &scrape, NULL ) )
     {
         return 1;
     }
@@ -629,20 +748,20 @@ int tr_trackerScrape( tr_torrent_t * tor, int * seeders, int * leechers )
     return 0;
 }
 
-int tr_trackerSeeders( tr_torrent_t * tor)
+int tr_trackerSeeders( tr_tracker_t * tc )
 {
-	if (tor->status != TR_STATUS_PAUSE)
-	{
-		return (tor->tracker)->seeders;
-	}
-	return 0;
+    if( !tc )
+    {
+        return -1;
+    }
+    return tc->seeders;
 }
 
-int tr_trackerLeechers( tr_torrent_t * tor)
+int tr_trackerLeechers( tr_tracker_t * tc )
 {
-	if (tor->status != TR_STATUS_PAUSE) 
-	{
-		return (tor->tracker)->leechers;
-	}
-	return 0;
+    if( !tc )
+    {
+        return -1;
+    }
+    return tc->leechers;
 }
